/* All content(c) 2016 - 2017 DigiPen(USA) Corporation, all rights reserved. */
/*****************************************************************************/
/*!
\file water.frag
\author Connor Deakin
\par E-mail: connortdeakin\@gmail.com
\par Project: X-Craft
\brief
  This is the shader that is responsible for making the water look the way it
  does. When I first made this, I was in a bit of a rush since we were only a
  week or two away from shipping. For that reason, I was not able to comment
  this. Four months later and I am here commenting it. Appologies if there are
  any inconsistencies or errors in my explanations.
*/
/*****************************************************************************/
#version 330 core

// Our Texture coordinate.
in vec2 texCoord;

// The current program time.
uniform float time = 0.0f;
// The current position of the game's camera.
uniform vec2 cameraPosition = vec2(1.0, 1.0);
// The frequency of the noise.
uniform float frequency = 0.5;
// The speed at which noise moves.
uniform float speedMultiple = 0.5;
// The number of world units displayed horizontally and vertically on the
// screen.
uniform vec2 worldSize = vec2(15.0, 15.0);
// The number of layers or colors the water will be composed of.
uniform int layers = 10;
// The color used for peaks in the water.
uniform vec4 peakColor = vec4(0.0, 0, 0, 1.0);
// The color used for troughs in the water.
uniform vec4 troughColor = vec4(0.0, 0.0, 0, 1.0);

// The resulting color.
out vec4 color;

// The code within this codeblock is simply perlin noise. I won't go through
// any explanations since there are already plenty of resources on this
// technique. If you are interested in learning about this or would like to
// revisit, I highly reccomend "The Book of Shaders". Here is a link to the
// article. https://thebookofshaders.com/11/. The previous articles highlight
// on other techniques used in the article linked.
// START CODE BLOCK //
float rand(vec2 seed)
{
  float dot_result = dot(seed.xy, vec2(12.9898,72.233));
  return fract(sin(dot_result) * 43758.5453123);
}
float noise(vec2 seed)
{
  vec2 i = floor(seed);
  vec2 f = fract(seed);
  float tl = rand(i);
  float tr = rand(i + vec2(1.0, 0.0));
  float bl = rand(i + vec2(0.0, 1.0));
  float br = rand(i + vec2(1.0, 1.0));
  vec2 t = smoothstep(0.0, 1.0, f);
  return mix(tl, tr, t.x) +
    (bl - tl) * (1.0 - t.x) * t.y +
    (br - tr) * t.x * t.y;
}
// END CODE BLOCK //

void main()
{
  // The uv will start with values including and between 0 and 1
  // We will use these texture coordinates in order to simulate the water
  // surface.
  vec2 uv = texCoord;
  // Our first step is to find the seed we will be using to create our noise.
  // We'll split this into a few parts to make it easy to digest.
  /*
  1. uv * worldSize
    Remember, the UV coordinates are values in the range [0, 1].
    What we are doing here is scaling the current uv coordinates by the number
    of world units that are displayed on the screen in both the horizontal and
    vertical directions. If we use the default values for worldSize (15, 15)
    (Meaning there are 15 world units across the width of the screen and 15
    world units across the height of the screen), we will get resulting values
    that will be in the range of [0, 15].
  2. cameraPosition + (1.)
    Next, we need to take the cameras current position into account. Since our
    UV is properly scaled to the world size, we can add the camera's position
    so that our texture coordinates accurately repesent a location on the
    screen. A an example, let's take our calculation from part 1. If our current
    values are in the range of [0, 15] and our camera position is (1, 0), our
    new values are in the following ranges.
    [1, 16]
    [0, 15]
    That is exactly what we want. If the camera moves up one unit in the world,
    it moves up exactly one unit for our modified textur coordinate values
  3. frequency * (2.)
    This last part is the simplest. All we are doing here is scaling our range
    to be larger or smaller. If we want to see more troughs and peaks in the
    water, all we need to do is up the scale in the frequency. If we want to
    see less, we lower the frequency value.
  */
  vec2 position = frequency * (cameraPosition + uv * worldSize);
  // This will allow us to offset our noise maps when we create them.
  // The scale and frequency of the noise maps will not be affected, but their
  // positions will be.
  float time_offset = time * speedMultiple;
  // Noise value moving up and to the right
  float noise_value_1 = noise(position + time_offset);
  // Noise value moving down and to the left
  float noise_value_2 = noise(position - time_offset);
  // To get our final noise value, all we do is find the value directly in
  // between noise_value_1 and noise_value_2.
  float noise_value = mix(noise_value_1, noise_value_2, 0.5);

  // This part is really short, but it's where a majority of the style comes
  // into the shader. Like the section before, I am also going to break this
  // part down
  /*
  1. noise_value * float(layers);
    Currently, noise_value is garunteed to be in the range of [0, 1].
    By multiplying by layers, we shift that range to [0, layers].
  2. floor(noise_value)
    This is the most important part. Our noise_value was in the range of
    [0, layers], but after this operation, our noise_value can only be a finit
    number of values. For example, let's say layers has the value of 5. First
    our noise_value goes to the range [0, 5] during step 1. During step 2, our
    noise_value will become one of the following values: 0, 1, 2, 3, 4, or 5,
    though, the value of 5 is very unlikely. This is the exact reason why the
    shader looks like it is only composed of a few colors.
  3. noise_value / float(layers)
    All we are doing now is putting our noise value back in the range of
    [0, 1]. Extending from the previous example in step 2's explanation where
    we ended with the values: 0, 1, 2, 3, 4, and 5. After this step, every
    noise_value that goes through this shader will have one of the following
    values: 0.0, 0.2, 0.4, 0.6, 0.8, or 1.0. With that, we have our linear
    interpolation parameter.
  */
  noise_value = noise_value * float(layers);
  noise_value = floor(noise_value);
  noise_value = noise_value / float(layers);
  // Taking our noise_value, we perform a linear interpolation between the
  // two colors to find out what color our fragment will be.
  color = mix(peakColor, troughColor, noise_value);
